import numpy as np
import h5py
from pathlib import Path

import pickle
import codecs


class DotDict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        return self.__dict__.update(d)

class AndersonAccelerationObj(DotDict):
    """Stores the current state of the program. Currently just a dictionary."""

def remove_nones(d):
    """
    Recursively removes None types from dictionary for MATLAB compatibility
    """
    none_keys = []
    for k, v in d.items():
        if v is None:
            none_keys.append(k)
        if isinstance(v, dict):
            d[k] = remove_nones(v)

    for k in none_keys:
        del d[k]

    return d


def translate_type(d, from_, to_):
    """
    Translates values in dict types recursively
    """
    if isinstance(d, dict):
        for k, v in d.items():
            if type(v) == from_:
                d[k] = translate_type(v, from_, to_)

            if isinstance(v, list):
                for i, val in enumerate(v):
                    if type(val) == from_:
                        v[i] = translate_type(val, from_, to_)

    return to_(d)


def determine_numpy_type(x):
    try:
        # Try to make a numpy array
        return np.array(x).dtype
    except ValueError:
        # Data is ragged. Let's assume datatype homogeneity and get the first element.
        return determine_numpy_type(x[0])


def save(state, filename, vars_=None, debug=False, comp_default=0):
    """
    Save state in a hdf5 file.
    """
    
    if Path(filename).suffix=='':
      filename=filename+".h5"

    if debug:
        print(f"Saving {filename}")

    save_state = {k: v for k, v in state.items() if (vars_ and k in vars_) or not vars_}
    save_state = remove_nones(save_state)
    save_state = translate_type(save_state, DotDict, dict)

    def pickle_val_to_group(group, k, v):
        pickle_bytes = pickle.dumps(v, protocol=pickle.HIGHEST_PROTOCOL)
        b64 = codecs.encode(pickle_bytes, "base64").decode("utf-8")
        np_b64 = np.array(b64, dtype="S")
        string_dt = h5py.special_dtype(vlen=str)

        return group.create_dataset(k, data=np_b64, dtype=string_dt)

    def add_contents_to_group(group, contents):
        # if we can't deal with a particular datatype, in principle we can
        # pickle it and store the binary data.

        for k, v in contents.items():
            if debug:
                print("Saving key", k)
                print("With val of type", type(v))

            if k == "__class__":
                continue

            if isinstance(v, dict):
                new_group = group.create_group(k)
                add_contents_to_group(new_group, v)
                continue

            # list is inhomogeneous, save as dictionaries
            if isinstance(v, list):
                new_group = group.create_group(k)
                new_group.attrs["list_as_group"] = True
                list_dict = {str(i): e for i, e in enumerate(v)}
                add_contents_to_group(new_group, list_dict)
                continue

            dtype = determine_numpy_type(v)
            # deal with strings
            if dtype.char in ["U", "S"]:
                # only for UTF-8, (i.e, default python strings)
                dt = h5py.string_dtype(encoding="utf-8")

                if isinstance(v, str):
                    utf_v = v.encode("utf-8")
                elif isinstance(v, bytes):
                    utf_v = v
                else:
                    utf_v = [e.encode("utf-8") if isinstance(e, str) else e for e in v]

                dataset = group.create_dataset(k, data=utf_v, dtype=dt)
                dataset.attrs["is_string"] = True
                continue

            # let hdf5 deal with other types by default
            try:
                comp_kwargs = {"compression": "gzip", "compression_opts": comp_default}
                if np.isscalar(v):
                    comp_kwargs = {}
                dataset = group.create_dataset(k, data=v, **comp_kwargs)
            # if hdf5 failed to save, pickle it. (This breaks portability, so shouldn't happen!)
            except Exception as e:
                print("Unexpected fail to translate to HDF5, pickling instead.", k)
                print(v.dtype, e)
                # save arbitrary data by pickling it
                dataset = pickle_val_to_group(group, k, v)

    with h5py.File(f"{filename}", "w") as f:
        add_contents_to_group(f, save_state)

    if debug:
        print(f"Saved {filename}")


def load(filename, vars_=None, debug=False, aaobj=False):
    """
    Load state from hdf5 file
    """

    if Path(filename).suffix=='':
      filename=filename+".h5"

    if debug:
        print(f"Loading {filename}")

    def add_from_group(group, group_is_list=False, only_load_vars=None):
        if group_is_list:
            state = [None for _ in range(len(group.keys()))]
        else:
            state = {}

        for key, val in group.items():
            if only_load_vars and key not in only_load_vars:
                continue

            if group_is_list:
                key = int(key)

            if isinstance(val, h5py.Dataset):
                data = val[()]

                if val.attrs.get("is_string"):
                    if isinstance(data, bytes):
                        data = data.decode("utf-8")
                    else:
                        data = np.char.decode(data.astype("S"), encoding="utf-8")

                if val.attrs.get("is_pickled"):
                    data = pickle.loads(codecs.decode(data, "base64"))

                # convert numpy scalars to python scalars
                if np.isscalar(data) and type(data).__module__ == np.__name__:
                    data = data.item()

                state[key] = data

            if isinstance(val, h5py.Group):
                list_as_group = val.attrs.get("list_as_group")
                state[key] = add_from_group(val, group_is_list=list_as_group)

        return state

    with h5py.File(f"{filename}", "r") as f:
        load_state = add_from_group(f, only_load_vars=vars_)

    load_state = translate_type(load_state, dict, DotDict)
    return AndersonAccelerationObj(load_state) if aaobj else load_state

